// =================================================================================================
// CopySourceAsHtml
// 
// Copyright  2006 J.T. Leigh & Associates Inc.  All rights reserved.
// 
// Microsoft and Visual Studio are either trademarks or registered trademarks 
// Corporation.
// 
// THIS SOFTWARE IS PROVIDED "AS-IS," WITHOUT WARRANTY OF ANY KIND, AND ANY USE OF THIS SOFTWARE 
// PRODUCT IS AT YOUR OWN RISK.  TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, J.T. LEIGH &
// ASSOCIATES AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES AND CONDITIONS, EITHER EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES AND CONDITIONS OF MERCHANTABILITY AND/OR 
// FITNESS FOR A PARTICULAR PURPOSE, TITLE, AND NON-INFRINGEMENT, WITH REGARD TO THE SOFTWARE. 
// =================================================================================================

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Text;
using System.Text.RegularExpressions;
using System.Web;
using JTLeigh.Tools.CopySourceAsHtml.Properties;

namespace JTLeigh.Tools.CopySourceAsHtml
{
   
    public sealed class Copier : RtfParser
    {
        private static Regex colorPattern = new Regex(@"\\red(?<r>\d+)\\green(?<g>\d+)\\blue(?<b>\d+)", RegexOptions.Compiled | RegexOptions.ExplicitCapture);

        private ColorableItemStyle plainTextStyle;
        private ColorableItemStyle lineNumberStyle;
        private Dictionary<int, HtmlColor> htmlColors;
        private HtmlFile htmlFile;
        private Encoding encoding;

        public Copier()
        {
            this.plainTextStyle = new ColorableItemStyle("black","white",false);
            this.lineNumberStyle = new ColorableItemStyle("silver", "white", false);
            Settings.Default.NumberLines = false;
            Settings.Default.IncludeRtf = false;
            Settings.Default.RemoveIndentation = true;
            Settings.Default.WrapWords = false;
        }
         
        public Copier(ColorableItemStyle plainTextStyle, ColorableItemStyle lineNumberStyle)
        {
            this.plainTextStyle = plainTextStyle;
            this.lineNumberStyle = lineNumberStyle;
            Settings.Default.NumberLines = false;
        }

        public string CopyAsHtml(string rtf)
        {

            int undentSpaces;
            StringBuilder html;
            HtmlStyleSheet styleSheet = null;
            HtmlStyle fileStyle;
            HtmlStyle lineStyle;
            HtmlStyle blockStyle;
            string content;
            int lineIndex;
            int lineNumber;
            int undentSpacesRemaining;
            bool allowLeadingSpace;
            string foregroundColor;
            string backgroundColor;

            this.htmlColors = new Dictionary<int, HtmlColor>();
            this.htmlFile = new HtmlFile(Settings.Default.WrapWords, Settings.Default.TabSize);
            this.encoding = Encoding.Default;

            ParseRtf(rtf);

            if (Settings.Default.RemoveIndentation)
            {
                undentSpaces = Int32.MaxValue;
                foreach (HtmlLine line in this.htmlFile.Lines)
                {
                    if ((line.HasContent) && (line.LeadingSpaces < undentSpaces))
                    {
                        undentSpaces = line.LeadingSpaces;
                    }
                }
            }
            else
            {
                undentSpaces = 0;
            }

            html = new StringBuilder();

            fileStyle = new HtmlStyle();
            fileStyle.AppendFormat("font-family: {0};", Settings.Default.FontList);
            fileStyle.AppendFormat("font-size: {0}pt;", Settings.Default.FontSize);
            if (this.plainTextStyle != null)
            {
                fileStyle.Append(this.plainTextStyle.ToString());
            }
            if ((Settings.Default.FileStyle != null) && (Settings.Default.FileStyle.Length != 0))
            {
                fileStyle.Append(Settings.Default.FileStyle);
            }

            lineStyle = new HtmlStyle();
            lineStyle.Append("margin: 0px;");
            if ((Settings.Default.LineStyle != null) && (Settings.Default.LineStyle.Length != 0))
            {
                lineStyle.Append(Settings.Default.LineStyle);
            }

            if (!Settings.Default.EmbedStyles)
            {
                styleSheet = new HtmlStyleSheet();
                if (fileStyle.Length > 0)
                {
                    styleSheet.AddStyle("cf", fileStyle);
                }
                if (lineStyle.Length > 0)
                {
                    styleSheet.AddStyle("cl", lineStyle);
                }
                if (Settings.Default.NumberLines && (this.lineNumberStyle != null) && (this.lineNumberStyle.Length > 0))
                {
                    styleSheet.AddStyle("cln", this.lineNumberStyle);
                }
            }

            html.Append("<div");
            if (fileStyle.Length > 0)
            {
                if (Settings.Default.EmbedStyles)
                {
                    html.AppendFormat(" style=\"{0}\"", fileStyle.ToString());
                }
                else
                {
                    html.Append(" class=\"cf\"");
                }
            }
            html.Append(">");
            if (!Settings.Default.StripLineBreaks)
            {
                html.Append("\n");
            }

            lineIndex = 1;
            lineNumber = Settings.Default.NumberLinesFrom;
            foreach (HtmlLine line in this.htmlFile.Lines)
            {
                undentSpacesRemaining = undentSpaces;
                if (line.HasContent || (lineIndex != this.htmlFile.Lines.Count))
                {

                    if (Settings.Default.WrapWords)
                    {
                        html.Append("<p");
                    }
                    else
                    {
                        html.Append("<pre");
                    }
                    if (lineStyle.Length > 0)
                    {
                        if (Settings.Default.EmbedStyles)
                        {
                            html.AppendFormat(" style=\"{0}\"", lineStyle.ToString());
                        }
                        else
                        {
                            html.Append(" class=\"cl\"");
                        }
                    }
                    html.Append(">");

                    if (Settings.Default.NumberLines)
                    {
                        if ((this.lineNumberStyle != null) && (this.lineNumberStyle.Length > 0))
                        {
                            html.Append("<span");
                            if (Settings.Default.EmbedStyles)
                            {
                                html.AppendFormat(" style=\"{0}\"", this.lineNumberStyle.ToString());
                            }
                            else
                            {
                                html.Append(" class=\"cln\"");
                            }
                            html.Append(">");
                        }
                        html.Append(lineNumber.ToString(NumberFormatInfo.InvariantInfo).PadLeft(5).Replace(" ", "&nbsp;"));
                        if (this.lineNumberStyle != null)
                        {
                            html.Append("</span>");
                        }
                        html.Append("&nbsp;");
                    }

                    if (line.HasContent)
                    {
                        allowLeadingSpace = false;
                        foreach (HtmlBlock block in line.Blocks)
                        {
                            if (block.HasContent || (block.LeadingSpaces > 0))
                            {
                                content = block.GetHtml(allowLeadingSpace, undentSpacesRemaining);
                                undentSpacesRemaining -= block.LeadingSpaces;
                                if (undentSpacesRemaining < 0)
                                {
                                    undentSpacesRemaining = 0;
                                }
                                if (content.Length > 0)
                                {
                                    allowLeadingSpace = true;
                                    blockStyle = new HtmlStyle();
                                    if (this.htmlColors.ContainsKey(block.ForegroundColor))
                                    {
                                        foregroundColor = this.htmlColors[block.ForegroundColor].ToString();
                                        if (foregroundColor != this.plainTextStyle.ForegroundColor)
                                        {
                                            blockStyle.AppendFormat("color: {0};", foregroundColor);
                                        }
                                    }
                                    if (this.htmlColors.ContainsKey(block.BackgroundColor))
                                    {
                                        backgroundColor = this.htmlColors[block.BackgroundColor].ToString();
                                        if (backgroundColor != this.plainTextStyle.BackgroundColor)
                                        {
                                            blockStyle.AppendFormat("background: {0};", backgroundColor);
                                        }
                                    }
                                    if (block.Bold != this.plainTextStyle.Bold)
                                    {
                                        if (block.Bold)
                                        {
                                            blockStyle.Append("font-weight: bold;");
                                        }
                                        else
                                        {
                                            blockStyle.Append("font-weight: normal;");
                                        }
                                    }
                                    if ((Settings.Default.BlockStyle != null) && (Settings.Default.BlockStyle.Length > 0))
                                    {
                                        blockStyle.Append(Settings.Default.BlockStyle);
                                    }
                                    if (blockStyle.Length > 0)
                                    {
                                        html.Append("<span");
                                        if (Settings.Default.EmbedStyles)
                                        {
                                            html.AppendFormat(" style=\"{0}\"", blockStyle.ToString());
                                        }
                                        else
                                        {
                                            html.AppendFormat(" class=\"{0}\"", styleSheet.ReuseStyle("cb", blockStyle));
                                        }
                                        html.Append(">");
                                    }
                                    html.Append(content);
                                    if (blockStyle.Length > 0)
                                    {
                                        html.Append("</span>");
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        if (!Settings.Default.NumberLines)
                        {
                            html.Append("&nbsp;");
                        }
                    }

                    if (Settings.Default.WrapWords)
                    {
                        html.Append("</p>");
                    }
                    else
                    {
                        html.Append("</pre>");
                    }
                    if (!Settings.Default.StripLineBreaks)
                    {
                        html.Append("\n");
                    }
                }

                lineIndex++;
                lineNumber++;

            }

            html.Append("</div>");
            if (!Settings.Default.StripLineBreaks)
            {
                html.Append("\n");
            }

            if (!Settings.Default.EmbedStyles)
            {
                html.Insert(0, styleSheet.GetHtml(Settings.Default.StripLineBreaks));
            }

            if (Settings.Default.IncludeRtf)
            {
                html.Insert(0, String.Format("<!--\n{0}\n-->\n", Tools.HtmlEncode(rtf)));
            }

            return html.ToString();

        }

        protected override int ParseControlWord(string rtf, int next, string controlWord, bool hasParameter, long parameter)
        {
            switch (controlWord)
            {
                case "ansicpg":
                    if (hasParameter)
                    {
                        this.encoding = Encoding.GetEncoding((int)parameter);
                    }
                    return 0;
                case "fonttbl":
                    return ProcessFontTable(rtf, next);
                case "colortbl":
                    return ProcessColorTable(rtf, next);
                case "par":
                    this.htmlFile.StartNewLine();
                    return 0;
                case "tab":
                    AppendTab();
                    return 0;
                case "u":
                    if (hasParameter)
                    {
                        if ((char)parameter == '\t')
                        {
                            AppendTab();
                        }
                        else
                        {
                            this.htmlFile.AppendEncoded((char)parameter);
                        }
                    }
                    return 1;
                case "cf":
                    if (hasParameter)
                    {
                        this.htmlFile.ForegroundColor = (int)parameter;
                    }
                    return 0;
                case "cb":
                    if (hasParameter)
                    {
                        this.htmlFile.BackgroundColor = (int)parameter;
                    }
                    return 0;
                case "b":
                    if (hasParameter && (parameter == 0))
                    {
                        this.htmlFile.Bold = false;
                    }
                    else
                    {
                        this.htmlFile.Bold = true;
                    }
                    return 0;
                default:
                    return 0;
            }
        }

        protected override int ParseControlSymbol(string rtf, int next, char controlSymbol)
        {

            switch (controlSymbol)
            {
                case '\\':
                case '{':
                case '}':
                    this.htmlFile.AppendEncoded(controlSymbol);
                    return 0;
                case '\'':
                    return ProcessHexEscapedCharacters(rtf, next); // -2 so next begins with \'
                default:
                    break;
            }
            return 0;
        }

        protected override int ParseOpeningBrace(string rtf, int next)
        {
            this.htmlFile.PushBlock();
            return 0;
        }

        protected override int ParseClosingBrace(string rtf, int next)
        {
            this.htmlFile.PopBlock();
            return 0;
        }

        protected override int ParseLiteralText(string rtf, int next, char literalText)
        {
            this.htmlFile.AppendEncoded(literalText);
            return 0;
        }

        private int ProcessHexEscapedCharacters(string rtf, int next)
        {

            int first;
            int current;
            int last;
            byte byteValue;
            List<byte> byteList;
            byte[] byteArray;
            char[] charArray;

            first = next - 2;
            current = first;
            last = rtf.Length;
            byteList = new List<byte>();

            while (((current + 1) < last) && (rtf[current] == '\\') && (rtf[(current + 1)] == '\''))
            {
                current += 2;
                if (((current + 1) < last) && char.IsLetterOrDigit(rtf, current) && char.IsLetterOrDigit(rtf, (current + 1)))
                {
                    byteList.Add(Byte.Parse(rtf.Substring(current, 2), NumberStyles.HexNumber));
                    current += 2;
                }
                else if ((current < last) && char.IsLetterOrDigit(rtf, current))
                {
                    byteValue = Byte.Parse(rtf.Substring(next, 1), NumberStyles.HexNumber);
                    if (byteValue == 0)
                    {
                        // map "\'0" to a space
                        byteValue = 32;
                    }
                    byteList.Add(byteValue);
                    current++;
                }
                else
                {
                    break;
                }
            }
            if (byteList.Count > 0)
            {
                byteArray = byteList.ToArray();
                charArray = this.encoding.GetChars(byteArray);
                foreach (char charValue in charArray)
                {
                    switch (charValue)
                    {
                        case '\t':
                            AppendTab();
                            break;
                        case ' ':
                            this.htmlFile.AppendUnencoded(' ');
                            break;
                        default:
                            this.htmlFile.AppendEncoded(charValue);
                            break;
                    }
                }
            }
            return (current - next);

        }

        private int ProcessFontTable(string rtf, int next)
        {

            int last;
            int index;
            int depth;

            last = rtf.Length;
            index = next;
            depth = 1;
            while ((index < last) && (depth > 0))
            {
                switch (rtf[index])
                {
                    case '{':
                        depth++;
                        break;
                    case '}':
                        depth--;
                        break;
                }
                index++;
            }
            return (index - next);

        }

        private int ProcessColorTable(string rtf, int next)
        {

            Match colorMatch;
            int colorID;
            int last;
            int index;
            int mark;
            int r;
            int g;
            int b;

            colorID = 0;
            last = rtf.Length;
            index = next;
            mark = index;
            while ((index < last) && (rtf[index] != '}'))
            {
                if (rtf[index] == ';')
                {
                    colorMatch = colorPattern.Match(rtf.Substring(mark, index - mark));
                    if (colorMatch.Success)
                    {
                        r = Convert.ToInt32(colorMatch.Groups["r"].Value);
                        g = Convert.ToInt32(colorMatch.Groups["g"].Value);
                        b = Convert.ToInt32(colorMatch.Groups["b"].Value);
                        this.htmlColors.Add(colorID, new HtmlColor(r, g, b));
                    }
                    colorID++;
                    mark = index + 1;
                }
                index++;
            }
            return (index - next);

        }

        private void AppendTab()
        {
            int space;

            for (space = 0; space < Settings.Default.TabSize; space++)
            {
                this.htmlFile.AppendUnencoded(' ');
            }

        }

    }

}
